from dataclasses import dataclass
import json
import os
from pathlib import Path
from typing import Iterable, Iterator, List, Literal, Mapping, Optional, TypedDict

import fastjsonschema
import tomlkit


HookName = Literal['build', 'lock']


config_schema = {
    'type': 'object',
    'properties': {
        'members': {
            'type': 'array',
            'items': {
                'type': 'string'
            }
        },
        'exclude': {
            'type': 'array',
            'items': {
                'type': 'string'
            }
        },
        'hooks': {
            'type': 'array',
            'items': {
                'type': 'string',
                'enum': ['build', 'lock']
            }
        }
    }
}


class MultiverseToml(TypedDict, total=False):
    members: List[str]
    exclude: List[str]
    hooks: List[HookName]


@dataclass
class WorkspaceConfiguration:
    root: Path
    config: MultiverseToml
    env: Mapping[str, str]

    @property
    def project_dirs(self) -> Iterable[Path]:
        def projects(globs: List[str]) -> Iterator[Path]:
            for dir_glob in globs:
                toml_glob = f'{dir_glob.rstrip(os.sep)}{os.sep}pyproject.toml'
                yield from self.root.glob(toml_glob)

        excluded = set(projects(self.config.get('exclude', [])))
        for included in projects(self.config.get('members', ['**'])):
            if included not in excluded:
                yield included.parent.resolve(True)
    
    def hooks_enabled(self, *hooks: HookName) -> bool:
        if bool(json.loads(self.env.get('MULTIVERSE_DISABLE_HOOKS', 'false'))):
            return False
        return all(hook in self.config.get('hooks', []) for hook in hooks)
    
    @staticmethod
    def find(
        path: Path,
        env: Mapping[str, str] = os.environ
    ) -> Optional['WorkspaceConfiguration']:
        if (path / 'multiverse.toml').is_file():
            return WorkspaceConfiguration.load(path / 'multiverse.toml', env)
        if str(path) == path.root:
            return None
        return WorkspaceConfiguration.find(path.parent, env)
    
    @staticmethod
    def load(path: Path, env: Mapping[str, str] = os.environ) -> 'WorkspaceConfiguration':
        file = path.read_text()
        if file.strip() == '':
            return WorkspaceConfiguration(path.parent, MultiverseToml({}), env)
        doc = tomlkit.parse(file).unwrap()
        validate = fastjsonschema.compile(config_schema)
        validate(doc)  # type:ignore
        return WorkspaceConfiguration(path.parent, MultiverseToml(**doc), env)



from pathlib import Path
from typing import Optional

from poetry.console.commands import show
from poetry.puzzle.provider import IncompatibleConstraintsError

from poetry_plugin_multiverse.commands.workspace import WorkspaceCommand
from poetry_plugin_multiverse.config import WorkspaceConfiguration
from poetry_plugin_multiverse.repositories import lock, locked_pool
from poetry_plugin_multiverse.root import root_env
from poetry_plugin_multiverse.workspace import Workspace


class ShowCommand(WorkspaceCommand):
    name = 'workspace show'
    description = 'List dependencies in the multiverse workspace.'

    arguments = show.ShowCommand.arguments
    options = show.ShowCommand.options
    aliases = show.ShowCommand.aliases
    usages = show.ShowCommand.usages
    commands = show.ShowCommand.commands

    def _workspace(self) -> Optional[Workspace]:
        path = self.io.input.option('directory')
        if config := WorkspaceConfiguration.find(Path(path) if path else Path.cwd()):
            return Workspace(config)
        return None

    def handle(self) -> int:
        workspace = self._workspace()
        if not workspace:
            self.io.write_error('Unable to locate workspace root')
            return 1

        root = workspace.root
        try:
            return_code = lock(root, locked_pool(*workspace.projects, packages=list(workspace.packages), strict=True))
            if return_code != 0:
                self.io.write_error_line('<error>Failed to lock workspace!</>')
                return return_code
        except IncompatibleConstraintsError:
            self.io.write_error_line('<error>Failed to lock workspace!</>')
            return 1

        show_command = show.ShowCommand()
        show_command.set_application(self.application)
        show_command.set_poetry(root)
        show_command.set_env(root_env(root))
        return show_command.run(self.io)



from contextlib import contextmanager
import traceback
from typing import Iterator

from cleo.io.io import IO
from cleo.io.outputs.output import Verbosity


@contextmanager
def error_boundary(io: IO) -> Iterator[None]:
    try:
        yield
    except Exception as ex:
        io.write_error_line([
            'An error occured in the Multiverse plugin:',
            str(ex)
        ])
        io.write_error_line(
            traceback.format_exc(),
            verbosity=Verbosity.VERBOSE  # type:ignore  # Poetry type is wrong
        )



import os
from pathlib import Path
import platform

from poetry.config.config import Config
from poetry.core.constraints.version.parser import parse_single_constraint
from poetry.core.constraints.version.util import constraint_regions
from poetry.core.packages.project_package import ProjectPackage
from poetry.core.version.pep440.version import PEP440Version
from poetry.factory import Factory
from poetry.packages.locker import Locker
from poetry.poetry import Poetry
from poetry.utils.env.base_env import Env
from poetry.utils.env.mock_env import MockEnv
from poetry.utils.env.null_env import NullEnv


def root_project(*projects: Poetry, path: Path) -> Poetry:
    aggregate_project = ProjectPackage('workspace', '0.0.0')
    local_config = {}

    py_versions = parse_single_constraint('*')
    for project in projects:
        py_versions = py_versions.intersect(project.package.python_constraint)
    aggregate_project.python_versions = str(py_versions)

    for project in projects:
        for dep in project.package.all_requires:
            aggregate_project.add_dependency(dep)

    pyproject = path / 'workspace.toml'
    pyproject.write_text(Factory.create_pyproject_from_package(aggregate_project).as_string())
    return Poetry(
        file=pyproject,
        local_config=local_config,
        package=aggregate_project,
        locker=Locker(path / 'workspace.lock', local_config),
        config=Config.create()
    )


def root_env(project: Poetry) -> Env:
    null_env = NullEnv()

    def mock_env(version: PEP440Version) -> MockEnv:
        return MockEnv(
            (version.major, version.minor or 0, version.patch or 0),
            python_implementation=null_env.python_implementation,
            platform=null_env.platform,
            platform_machine=platform.machine(),
            os_name=os.name,
        )

    constraint = project.package.python_constraint
    if constraint.is_empty() or constraint.is_any():
        return null_env
    for region in constraint_regions([constraint]):
        if region.include_min and region.min:
            return mock_env(region.min)
        if region.include_max and region.max:
            return mock_env(region.max)
    return null_env



from contextlib import _GeneratorContextManager, contextmanager
from dataclasses import dataclass
from pathlib import Path
from queue import Queue
from threading import Thread
from typing import Callable, Iterator, Optional

from cleo.commands.command import Command
from cleo.io.io import IO

from poetry_plugin_multiverse.cli.progress import progress
from poetry_plugin_multiverse.cli.status import OutputQueue, StatusConfig, WorkspaceStatus
from poetry_plugin_multiverse.config import WorkspaceConfiguration
from poetry_plugin_multiverse.workspace import Workspace


@dataclass
class CliUtils:
    workspace: Workspace
    io: IO

    def progress(self, message: str) -> _GeneratorContextManager[None]:
        return progress(self.io, message)
    
    @contextmanager
    def status(self, header: Optional[str], action: str) -> Iterator[WorkspaceStatus]:
        q: Queue[Optional[Callable[[], None]]] = Queue()

        def process():
            while task := q.get():
                task()

        io_thread = Thread(target=process)
        try:
            io_thread.start()
            yield WorkspaceStatus(
                OutputQueue(q.put, self.io),
                self.workspace.projects,
                StatusConfig(header, StatusConfig.default_template(action))
            )
        finally:
            q.put(None)
            io_thread.join()


class WorkspaceCommand(Command):
    def _workspace(self) -> Optional[Workspace]:
        path = self.io.input.option('directory')
        if config := WorkspaceConfiguration.find(Path(path) if path else Path.cwd()):
            return Workspace(config)
        return None

    @property
    def cli(self) -> CliUtils:
        workspace = self._workspace()
        assert workspace is not None
        return CliUtils(workspace, self.io)

    def handle_workspace(self, workspace: Workspace) -> int:
        raise NotImplementedError
    
    def handle(self) -> int:
        if workspace := self._workspace():
            return self.handle_workspace(workspace)

        self.line_error('Unable to locate workspace root')
        return 1




from dataclasses import dataclass
from typing import Callable, List, Type
from cleo.commands.command import Command as CleoCommand
from cleo.events.console_event import ConsoleEvent
from cleo.events.console_events import COMMAND, TERMINATE
from cleo.events.event import Event
from cleo.events.event_dispatcher import EventDispatcher
from poetry.console.application import Application
from poetry.plugins.application_plugin import ApplicationPlugin

from poetry_plugin_multiverse.commands.check import CheckCommand
from poetry_plugin_multiverse.commands.info import InfoCommand
from poetry_plugin_multiverse.commands.lock import LockCommand
from poetry_plugin_multiverse.commands.run import RunCommand
from poetry_plugin_multiverse.commands.show import ShowCommand
from poetry_plugin_multiverse.errors import error_boundary
from poetry_plugin_multiverse.hooks.build import PreBuildHook
from poetry_plugin_multiverse.hooks.hook import Hook, HookContext
from poetry_plugin_multiverse.hooks.lock import PreLockHook, PostLockHook


@dataclass
class PluginConfig:
    commands: List[Type[CleoCommand]]
    pre_hooks: List[Callable[[], Hook]]
    post_hooks: List[Callable[[], Hook]]

    def configure(self, application: Application):
        for cmd in self.commands:
            assert cmd.name is not None
            application.command_loader.register_factory(cmd.name, cmd)

        if application.event_dispatcher:
            application.event_dispatcher.add_listener(COMMAND, self._before_command)
            application.event_dispatcher.add_listener(TERMINATE, self._after_command)
    
    def _before_command(self, event: Event, kind: str, dispatcher: EventDispatcher) -> None:
        if isinstance(event, ConsoleEvent):
            with error_boundary(event.io):
                if context := HookContext.create(event):
                    context.run(*self.pre_hooks)
    
    def _after_command(self, event: Event, kind: str, dispatcher: EventDispatcher) -> None:
        if isinstance(event, ConsoleEvent):
            with error_boundary(event.io):
                if context := HookContext.create(event):
                    context.run(*self.post_hooks)


class MultiversePlugin(ApplicationPlugin):
    def activate(self, application: Application):
        PluginConfig(
            commands=[
                CheckCommand,
                InfoCommand,
                LockCommand,
                RunCommand,
                ShowCommand
            ],
            pre_hooks=[
                PreBuildHook,
                PreLockHook
            ],
            post_hooks=[
                PostLockHook
            ]
        ).configure(application)



from collections import defaultdict
from dataclasses import dataclass
from typing import Iterator
from poetry.core.packages.package import Package
from poetry.poetry import Poetry

from poetry_plugin_multiverse.dependencies import Record


@dataclass
class Packages:
    records: dict[str, list[Record[Package]]]

    @staticmethod
    def from_projects(*projects: Poetry) -> 'Packages':
        deps: dict[str, list[Record[Package]]] = defaultdict(list)
        for project in projects:
            for dep in project.locker.locked_repository().packages:
                deps[dep.complete_name].append(Record(project, dep))
        return Packages(deps)

    def __iter__(self) -> Iterator[Package]:
        for record in self.records.values():
            for package in record:
                yield package.value



import os
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Iterable, Mapping, Optional

from poetry.factory import Factory
from poetry.poetry import Poetry

from poetry_plugin_multiverse.config import WorkspaceConfiguration
from poetry_plugin_multiverse.dependencies import Dependencies
from poetry_plugin_multiverse.packages import Packages
from poetry_plugin_multiverse.repositories import PoolFactory
from poetry_plugin_multiverse.root import root_project


class Workspace:
    def __init__(
        self,
        config: WorkspaceConfiguration, *,
        disable_cache: bool = False
    ) -> None:
        self.config = config
        self.disable_cache = disable_cache
        self.root_project = TemporaryDirectory()

    @staticmethod
    def create(
        poetry: Poetry, *,
        env: Mapping[str, str] = os.environ
    ) -> Optional['Workspace']:
        if config := WorkspaceConfiguration.find(poetry.pyproject_path.parent, env):
            if poetry.pyproject_path.parent in config.project_dirs:
                return Workspace(
                    config,
                    disable_cache=poetry.disable_cache
                )
            return None
    
    @property
    def root(self) -> Poetry:
        project = root_project(
            *self.projects,
            path=Path(self.root_project.name)
        )
        if pool := PoolFactory().get():
            project.set_pool(pool)
        return project
    
    @property
    def projects(self) -> Iterable[Poetry]:
        for project in self.config.project_dirs:
            try:
                project = Factory().create_poetry(
                    project,
                    disable_cache=self.disable_cache
                )
                if pool := PoolFactory().get():
                    project.set_pool(pool)
                yield project
            except RuntimeError:
                print(f'Skipping pyproject.toml {project}')
    
    @property
    def dependencies(self) -> Dependencies:
        return Dependencies.from_projects(*self.projects)
    
    @property
    def packages(self) -> Packages:
        return Packages.from_projects(*self.projects)



from concurrent.futures import ThreadPoolExecutor, as_completed
from cleo.helpers import option
from poetry.console.commands.lock import LockCommand as PoetryLockCommand
from poetry.poetry import Poetry

from poetry_plugin_multiverse.commands.workspace import WorkspaceCommand
from poetry_plugin_multiverse.repositories import lock, locked_pool, workspace_pool
from poetry_plugin_multiverse.workspace import Workspace


def explicit_regenerate() -> bool:
    return any(option.name == 'regenerate' for option in PoetryLockCommand.options)


class LockCommand(WorkspaceCommand):
    name = 'workspace lock'
    description = 'Compute dependencies for the multiverse workspace.'

    options = [
        option(
            "no-update", None, "Deprecated: Reuse packages in project lock files."
        ),
        option(
            "regenerate", None, "Ignore packages in project lock files."
        )
    ] if explicit_regenerate() else [
        option(
            "no-update", None, "Reuse packages in project lock files."
        )
    ]

    help = """
The <info>workspace lock</info> command updates the <comment>poetry.lock</>
file for each project in the workspace such that dependencies required by
multiple projects are locked to the same version.

<info>poetry workspace lock</info>
"""

    def handle_workspace(self, workspace: Workspace) -> int:
        return_code = 0
        root = workspace.root

        with self.cli.progress('Locking workspace...'):
            should_update = self.option('regenerate') if explicit_regenerate() else \
                not self.option('no-update')
            pool = workspace_pool(*workspace.projects) if should_update else \
                locked_pool(*workspace.projects, packages=list(workspace.packages), strict=True)
            return_code = lock(root, pool)
            if return_code != 0:
                self.io.write_error_line('<error>Failed to lock workspace!</>')
                return return_code
        
        with self.cli.status(None, 'Updating') as status:
            def run(project: Poetry) -> int:
                status(project).update('Locking...')
                return lock(
                    project,
                    locked_pool(
                        project,
                        packages=root.locker.locked_repository().packages,
                        strict=True
                    )
                )

            with ThreadPoolExecutor() as executor:
                jobs = {
                    executor.submit(lambda: run(project)): project
                    for project in workspace.projects
                }
                for future in as_completed(jobs):
                    project = jobs[future]
                    result = status(project).complete(future)
                    return_code = max(return_code, result if result is not None else 1)
        
        return return_code



from typing import Iterable, Optional
from poetry.core.packages.dependency import Dependency
from poetry.core.packages.path_dependency import PathDependency
from poetry.poetry import Poetry


class PublishedDependencies:
    def __init__(self, projects: Iterable[Poetry]) -> None:
        self.projects = {
            project.pyproject_path.parent.resolve(): project
            for project in projects
        }

    def __call__(self, dep: Dependency) -> Optional[Dependency]:
        if isinstance(dep, PathDependency):
            if project := self.projects.get(dep.full_path):
                return Dependency(
                    dep.name,
                    f'^{project.package.version}',
                    optional=dep.is_optional(),
                    groups=dep.groups,
                    extras=dep.extras
                )
        return None
    
    def patch_project(self, poetry: Poetry):
        for group_name in poetry.package.dependency_group_names(True):
            group = poetry.package.dependency_group(group_name)
            dependencies = list(group.dependencies)
            for dep in dependencies:
                if published := self(dep):
                    group.remove_dependency(dep.name)
                    group.add_dependency(published)
